#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2007-2008 Christian Dannie Storgaard
#
# AUTHOR:
# Christian Dannie Storgaard <cybolic@gmail.com>
#
# Cache provides a simple dict like object.
# The cache that is queried will always follow the environment variable WINEPREFIX,
# meaning that if you set a value and then change WINEPREFIX, the value will not be
# found until you set WINEPREFIX back to its previous value.
#
# Any value set in the cache has a timecode included that corresponds to the modified
# time of system.reg, userdef.reg and user.reg in the matching WINEPREFIX.
# If the timecode on a set value in the cache does not match the current timecode of
# the .reg files, the value/key will be reported as not set.
# This way you can always be sure that any value in the cache is up to date with the
# registry database.

import os, cPickle, copy, hashlib, re
from threading import Lock

pickle = cPickle

class Cache:
    def __init__(self):
        self.__version = 0.1
        self.__lock = Lock()
        self.__dict = {}
        self.__currentbottle = None
        self.__dict[None] = [0, {}]
        #self.__read_file_cache()
    
    def __getitem__(self, key):
        self.__lock.acquire()
        if self.__contains__(key):
            value = self.__dict[self.__currentbottle][1][key]
        else:
            raise KeyError, key
        retvalue = value
        self.__lock.release()
        return retvalue
    
    def __setitem__(self, key, value, update=True):
        if type(value) == dict:
            self.__dict[self.__currentbottle][1][key] = value.copy()
        else:
            self.__dict[self.__currentbottle][1][key] = value
        if update:
            self.changed()
    
    def __delitem__(self, key):
        self.__lock.acquire()
        try:
            del self.__dict[self.__currentbottle][1][key]
        finally:
            self.__lock.release()
    
    def __contains__(self, key):
        self.__read_file_cache()
        self.__lock.acquire()
        retvalue = key in self.__dict[self.__currentbottle][1] and self.__dict[self.__currentbottle][0] == self.__get_update_time()
        self.__lock.release()
        return retvalue
    
    def save(self, dict):
        for key, value in dict.iteritems():
            self.__setitem__(key, value, update=False)
        self.changed()
    
    def current(self):
        # Checking if cache is up to date...
        self.__read_file_cache()
        return self.__dict[self.__currentbottle][0] == self.__get_update_time()
    
    def changed(self):
        self.__write_file_cache()
    
    def copy(self):
        self.__read_file_cache()
        self.__lock.acquire()
        retvalue = copy.deepcopy(self.__dict[self.__currentbottle][1])
        self.__lock.release()
        return retvalue
    
    def clear(self):
        try:
            os.remove("%s/.pythonwinecache" % self.__currentbottle)
        except OSError:
            pass
        self.__currentbottle = None
    
    def __get_update_time(self):
        # Read all .reg files, remove the timestamps from the content of the files (the regex) and match that to our cached md5 sum
        combined_reg = ""
        for regfile in ['system.reg', 'userdef.reg', 'user.reg']:
            if os.access("%s/%s" % (self.__currentbottle, regfile), os.W_OK):
                f = open("%s/%s" % (self.__currentbottle, regfile), 'r')
                combined_reg += re.sub(r'(?m)^(\[[^\]\n\r]+\]) \d+(\s+)?$',r'\1', '\n'.join(f.readlines()) )
                f.close()
        md5 = hashlib.md5()
        md5.update(combined_reg)
        return md5.hexdigest()
    
    
    def __read_file_cache(self):
        self.__lock.acquire()
        # Check if WINEPREFIX has changed and only update the cache if it has
        if self.__currentbottle != os.environ['WINEPREFIX']:
            self.__currentbottle = os.environ['WINEPREFIX']
            
            cachefilename = "%s/.pythonwinecache" % self.__currentbottle
            if os.access(cachefilename, os.R_OK):
                #print "\tCache file exists, reading..."
                self.__filecache_file = open(cachefilename, 'rb')
                try:
                    self.__dict[self.__currentbottle] = pickle.load(self.__filecache_file)
                    #print "\tCache file read. Timestamp is %s." % self.__dict[self.__currentbottle][0]
                except:
                    self.__dict[self.__currentbottle] = [0, {}]
                    #print "\tSomething went wrong, resetting cache."
                self.__filecache_file.close()
                if '__version__' not in self.__dict[self.__currentbottle][1] or self.__dict[self.__currentbottle][1]['__version__'] < self.__version:
                    self.__dict[self.__currentbottle] = [0, {}]
            else:
                #print "\tNo cache file found, using blank."
                self.__dict[self.__currentbottle] = [0, {}]
        self.__lock.release()
    
    def __write_file_cache(self):
        self.__lock.acquire()
        self.__dict[self.__currentbottle][0] = self.__get_update_time()
        self.__dict[self.__currentbottle][1]['__version__'] = self.__version
        self.__filecache_file = open("%s/.pythonwinecache" % self.__currentbottle, 'wb')
        pickle.dump(self.__dict[self.__currentbottle], self.__filecache_file)
        self.__filecache_file.close()
        self.__lock.release()
