#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2007-2008 Christian Dannie Storgaard
#
# AUTHOR:
# Christian Dannie Storgaard <cybolic@gmail.com>

import _registry
import util

import subprocess, os, sys, re, difflib, copy

import logging

class Programs:
    def __init__(self):
        self.registry = _registry.Registry()
        self.logging = logging.getLogger("python-wine.programs")
        self.debug = self.logging.debug
        self.info = self.logging.info
        self.error = self.logging.error
        self.warning = self.logging.warning
        self.critical = self.logging.critical
        self.IGNORE_PROGRAMS_CONTAINING = ['unins', 'unvise', 'unwise', 'remove']
        self.DOCUMENT_MATCHERS = ['help', 'readme', 'read me', 'documentation', 'what\'s new', ' manual', '.pdf', ' website', 'release note', 'visit ', ' support']
        # Program exe/icon filename overrides in regex format
        self.PROGRAM_OVERRIDES = {
            'IE6': {
                'search': ('displayname', r'(?i)^Microsoft Internet Explorer 6'),
                'replace': [
                    ('name', "Microsoft Internet Explorer 6"),
                    ('description', "Web browser from Microsoft"),
                    ('publisher', "Microsoft Corporation"),
                    ('programexe', ('uninstallstring', r'(?i).*IE6Maintenance ([a-z]):\\.*', r'\1:\\Program Files\\Internet Explorer\\iexplore.exe')),
                    ('path', ('uninstallstring', r'(?i).*IE6Maintenance ([a-z]):\\.*', r'\1:\\Program Files\\Internet Explorer\\')),
                    ('programicon', ('key', 'programexe'))
                ]
            },
            '7-Zip': {
                'search': ('displayname', r'(?i)^7-Zip '),
                'replace': [
                    ('name', "7-Zip"),
                    ('description', "A file archiver with a high compression ratio"),
                    ('programexe', ('uninstallstring', r'(?i)[\'"]?(.*?)\\Uninstall.exe', r'\1\\7zFM.exe')),
                    ('path', ('uninstallstring', r'(?i)[\'"]?(.*?)\\Uninstall.exe', r'\1\\')),
                    ('programicon', ('key', 'programexe'))
                ]
            },
            'Fallout 3': {
                'search': ('displayname', '(?i)fallout ?3'),
                'replace': [
                    ('name', "Fallout 3"),
                    ('description', "A post-nuclear first-person role-playing game"),
                    ('publisher', "Bethesda Softworks Inc."),
                    ('programexe', ('installlocation', r'(?i)(.*)', r'\1\\FalloutLauncher.exe')),
                    ('icon', ('installlocation', r'(?i)(.*)', r'\1\\FalloutLauncher-MCE.png')),
                    ('path', ('key', 'installlocation'))
                ]
            }
        }
    
    def listFromReg(self):
        registry = self.registry.get('HKEY_LOCAL_MACHINE\\Software\\Microsoft\\Windows\\CurrentVersion\\Uninstall')
        programs = {}
        for program in registry.keys():
            if "VineyardName" in registry[program]:
                programname = registry[program]["VineyardName"]
                
            elif "DisplayName" in registry[program]:
                programname = registry[program]["DisplayName"]
            else:
                self.debug("Skipping program-entry \"%s\"" % program)
                continue
            """ Don't overwrite entrie with the same same """
            if programname in programs:
                programname = "%s (%s)" % (programname, program)
            programs[programname] = registry[program]
            programs[programname]['_registrykey'] = program
            self.debug("Adding program \"%s\"" % programname)
            if "UninstallString" in registry[program]:
                programs[programname]["UninstallString"] = re.sub(r'((^\\")|(\\"$))','', registry[program]["UninstallString"])
        return programs

    def listFromMenu(self):
        self.debug("Looking through menu entries made by wine...")
        output = subprocess.Popen(["find", "%s/../.local/share/applications/" % os.environ['WINEPREFIX'], "-iwholename", '*/wine/*.desktop'], stdout=subprocess.PIPE, stderr=subprocess.PIPE).communicate()[0]
        desktoplist = filter(len, output.split("\n"))
        
        self.debug("Filtering the found list and adding their info to the list...")
        desktopprograms = {}
        for path in desktoplist:
            desktopfile = open(path).readlines()
            name = command = icon = comment = None
        
            for line in desktopfile:
                line = line.replace("\n", '')
                if line.startswith("Name="):
                    name = '='.join(line.split('=')[1:])
                elif line.startswith("Exec="):
                    command = '='.join(line.split('=')[1:])
                elif line.startswith("Icon="):
                    icon = '='.join(line.split('=')[1:])
                elif line.startswith("Comment="):
                    comment = '='.join(line.split('=')[1:])
            
            prefix = re.match(r'env WINEPREFIX=["\']?(.*?)["\']? wine', command)
            if prefix:
                if util.wintounix(prefix.groups()[0]) == os.path.normpath(os.environ['WINEPREFIX']):
                    desktopprograms[name.lower()] = {"Name": name, "Command": command, "Icon": icon, "Comment": comment, "Path": os.path.dirname(path).split('/')[:-1]}
                else:
                    self.debug("Skipping menu entry for \"%s\", it doesn't seem to belong to this configuration." % name)
        self.debug("Done. Returning list of menu entries.")
        return desktopprograms
    
    def list(self, include_menu_only_entries=False):
        programsFromReg = self.listFromReg()
        programsFromMenuRaw = self.listFromMenu()
        # Filter the programsFromMenu dict for items containing names like "Uninstall", "Help", etc.
        programsFromMenu = dict((k, v) for (k, v) in programsFromMenuRaw.iteritems() if self.__isValidProgramName(k))
        # Remove empty values from programsFromMenu's subitems
        programsFromMenu = dict((k, dict((sk, sv) for (sk, sv) in v.iteritems() if sv is not None)) for (k, v) in programsFromMenu.iteritems())
        programs = {}
        menuOnlyPrograms = copy.deepcopy(programsFromMenu)
        
        if len(programsFromReg.keys()):
            for regKey,regData in programsFromReg.iteritems():
                # Lowercase all keys in regData to ensure better matching
                regData = util.dictToCaseInsensitive(regData)
                # Remove empty values
                regData = dict((k, v) for (k, v) in regData.iteritems() if len(v))
                
                # Don't count programs without names
                if 'vineyardname' not in regData and 'displayname' not in regData:
                    continue
                
                if 'vineyardname' in regData:
                    full_name = name = regData['vineyardname']
                else:
                    full_name = regData['displayname']
                    name = self.__returnValidProgramName(full_name, clean=False)
                
                program = {
                    'name': name,
                    '_registrykey': regData['_registrykey']
                }
                    
                for overridename,override in self.PROGRAM_OVERRIDES.iteritems():
                    if 'search' in override:
                        key, search = override['search']
                        if key in regData and re.search(search, regData[key]):
                            self.debug("Found override \"%s\" for \"%s\"" % (overridename, full_name))
                            for pkey,rvalue in override['replace']:
                                if type(rvalue) == type(()):
                                    if len(rvalue) == 3:
                                        match_key, match_value, replace_value = rvalue
                                        self.debug("Checking the override rule \"%s\" matches \"%s\"" % (match_value, regData[match_key]))
                                        match_value_matchobj = re.search(match_value, regData[match_key])
                                        if match_value_matchobj:
                                            self.debug("Rule matches, applying override logic to key \"%s\"" % pkey)
                                            program[pkey] = match_value_matchobj.expand(replace_value)
                                    elif len(rvalue) == 2:
                                        if rvalue[0] == "key" and rvalue[1] in regData:
                                            program[pkey] = regData[rvalue[1]]
                                else:
                                    program[pkey] = rvalue
                                if pkey in ['path', 'programexe', 'programicon', 'icon'] and pkey in program:
                                    program[pkey] = util.wintounix(program[pkey])
                
                # Copy data from registry info
                if 'vineyardname' in regData:
                    program['name'] = regData['vineyardname']
                if 'vineyardicon' in regData:
                    program['icon'] = self.getIcon({'name': program['name'], 'icon': regData['vineyardicon']})
                if 'vineyardcommand' in regData:
                    program['programcommand'] = regData['vineyardcommand']
                if 'vineyarddescription' in regData:
                    program['description'] = regData['vineyarddescription']
                if 'vineyardterminal' in regData:
                    program['programterminal'] = regData['vineyardterminal']
                if 'vineyarduninstall' in regData:
                    program['uninstall'] = regData['vineyarduninstall']
                
                if 'displayversion' in regData:
                    program['version'] = regData['displayversion']
                else:
                    match = re.search('(?:(?:\w+) ?)+ \(([\w\d\-\.]*)\)', full_name)
                    if match:
                        program['version'] = match.groups()[0]
                    else:
                        match = re.search(' [vV]?\.?((?:\d+\.?)*?)$', full_name)
                        if match:
                            program['version'] = match.groups()[0]
                if 'publisher' in regData:
                    program['publisher'] = regData['publisher']
                if 'uninstallstring' in regData:
                    program['uninstall'] = regData['uninstallstring']
                
                # Find icon from registry info and possibly executable
                if 'icon' not in program and 'displayicon' not in program and 'programicon' not in program:
                    if 'displayicon' in regData:
                        if regData['displayicon'].lower().split(',')[0].endswith('.exe'):
                            program['programicon'] = util.wintounix(regData['displayicon'].split(',')[0])
                            if 'programexe' not in program and self.__isValidProgramName(program['programicon']):
                                program['programexe'] = util.wintounix(regData['displayicon'].split(',')[0])
                        else:
                            program['icon'] = util.wintounix(regData['displayicon'].split(',')[0])
                    
                # Try to find the directory for the executable
                # We'll look for executable and icon files here
                if 'path' not in program:
                    self.debug("Finding path for program...")
                    if 'path' in regData:
                        program['path'] = util.wintounix(regData['path'])
                    elif 'basepath' in regData:
                        program['path'] = util.wintounix(regData['basepath'])
                    elif 'installlocation' in regData:
                        program['path'] = util.wintounix(regData['installlocation'])
                    elif 'uninstallstring' in regData:
                        # Don't use if uninstallstring doesn't contain an absolute path and only use the first executable, not its args
                        matches = [ match for match in re.findall(r'(?i)[\'"]?(\w\:\\.*?\.[a-z_]{3,4})[\'"]? ?', regData['uninstallstring']) if not match.lower().endswith('.dll') ]
                        if len(matches):
                            self.debug("Converting uninstall string to usable program path...")
                            matches = [ match for match in [ re.sub(r'(?i)(?<=\\)(%s).*?\.(log|txt|lst|exe)' % '|'.join(self.IGNORE_PROGRAMS_CONTAINING), 'fake_file_name_so_dirname_works', i) for i in matches ] ]
                            if len(matches):
                                self.debug("Found uninstallstring matches \"%s\" for the program path for \"%s\"" % (matches, program['name']))
                                match = matches[0]
                                program['path'] = '/'.join(util.wintounix(match).split('/')[:-1])
                                # If the innermost dir in the path is called something like "uninstall", use its parent 
                                if not self.__isValidProgramName( program['path'].split('/')[-1] ):
                                    program['path'] = '/'.join(program['path'].split('/')[:-1])
                
                # Find match for program['name'] in the menu files, but only if
                # we haven't already got all the information we can get from the menu file (from the PROGRAM_OVERRIDES f.x.)
                menuMatch = None
                if include_menu_only_entries or ('icon' not in program and 'programicon' not in program and 'description' not in program and 'programexe' not in program):
                    self.debug("Looking for a menu item matching \"%s\"" % program['name'])
                    menuMatchKey = None
                    if program['name'].lower() in programsFromMenu.keys():
                        # Matched directly
                        menuMatchKey = program['name'].lower()
                        menuMatch = programsFromMenu[menuMatchKey]
                        self.debug("Found matching program in programsFromMenu: \"%s\" (matched by \"%s\")" % (name, program['name'].lower()))
                    elif self.__returnValidProgramName(program['name']) in [ self.__returnValidProgramName(i) for i in programsFromMenu.keys() ]:
                        # Matched basic name (lowercase, strip anything in parenthesis)
                        for name in programsFromMenu.keys():
                            if self.__returnValidProgramName(program['name']) == self.__returnValidProgramName(name):
                                menuMatchKey = name
                                menuMatch = programsFromMenu[menuMatchKey]
                                self.debug("Found matching program in programsFromMenu: \"%s\" (matched by \"%s\")" % (name, self.__returnValidProgramName(name)))
                                break
                    else:
                        # Iterate through the Menu entries, sorted by length (shortest first)
                        simplestName = self.__returnValidProgramName(program['name'].split(' ')[0])
                        sortedMenuKeys = sorted(programsFromMenu.keys(), cmp=lambda a,b: len(a)-len(b))
                        # Prefer a match with a version number
                        for name in sortedMenuKeys:
                            if re.match('(%s( \w*?)?|(\w*? )?%s)[\- ][0-9\.\-]*' % (re.escape(simplestName), re.escape(simplestName)), name):
                                # Matched with a verson number
                                menuMatchKey = name
                                menuMatch = programsFromMenu[menuMatchKey]
                                self.debug("Found matching program in programsFromMenu: \"%s\" (matched by \"%s\")" % (name, simplestName))
                                break
                        if not menuMatch:
                            # Find a match to the simplest form of our program, possible stripping out a starting "play " from the name
                            for name in sortedMenuKeys:
                                if name.startswith(simplestName) or re.sub("^play ", '', name).startswith(simplestName):
                                    # Matched using the first part of the menu name and the basic name (see above)
                                    menuMatchKey = name
                                    menuMatch = programsFromMenu[menuMatchKey]
                                    self.debug("Found matching program in programsFromMenu: \"%s\" (matched by \"^play %s\")" % (name, simplestName))
                                    break
                # Done looking for a match in the menu files
                
                # Remove the match (if any found) from the list of menu-only menu entries
                if menuMatch:
                    try:
                        del menuOnlyPrograms[menuMatchKey]
                    except KeyError:
                        """ This program was already found.
                            This is probably a second install of the same program. """
                        pass
                
                # Search through the install directory for a possible executable and icon if either isn't found yet
                possible_exes = []
                if 'icon' not in program and 'programicon' not in program and 'programexe' not in program:
                    self.debug("Program so far: %s" % program)
                    self.debug("Searching for executables and icons...")
                    if 'path' in program and os.path.isdir(program['path']):
                        paths = [program['path']]
                        for filename in os.listdir(program['path']):
                            self.debug("\tChecking file in path (%s): %s..." % (program['path'], filename))
                            if filename.lower() == "bin":
                                self.debug("\t\tAdding dir \"%s/%s\" to exe search dir." % (program['path'], filename))
                                paths.append("%s/%s" % (program['path'], filename))
                        for path in paths:
                            for filename in filter(self.__isValidProgramName, os.listdir(path)):
                                self.debug("\tChecking if \"%s\" is an executable or an icon..." % filename)
                                if 'icon' not in program and 'programicon' not in program and filename.lower().split('.')[-1] in ['.ico', '.png', '.icon']:
                                    icon = "%s/%s" % (path, filename)
                                if filename.lower().endswith('.exe'):
                                    self.debug("\t\tExecutable \"%s\" found, adding to list." % filename)
                                    possible_exes.append("%s/%s" % (path, filename))
                    self.debug("Executables found: %s" % possible_exes)
                # Look for an executable in the list of executables in the program's path we just compiled
                if 'programcommand' not in program and 'programexe' not in program and len(possible_exes):
                    possible_exe = difflib.get_close_matches(program['name'], possible_exes, 3, 0.0)
                    if len(possible_exe):
                        program['programexe'] = possible_exe[0]
                
                # Copy usable values from the menuMatch into the program dict
                if menuMatch:
                    self.debug("Copying values from the found menu item...")
                    # Lowercase all keys in menuMatch to ensure better matching
                    menuData = util.dictToCaseInsensitive(menuMatch)
                    if 'icon' not in program and 'icon' in menuData:
                        program['icon'] = menuData['icon']
                    if 'comments' not in regData and 'comment' in menuData:
                        program['description'] = menuData['comment']
                    if 'programexe' not in regData and 'command' in menuData:
                        if menuData['command'].startswith('env WINEPREFIX=') or menuData['command'].startswith('wine '):
                            executable = self.__returnExecutableFromCommand(menuData['command'])
                            executable = util.wintounix(executable)
                            if self.__isValidProgramName(executable):
                                program['programcommand'] = menuData['command']
                                program['programexe'] = executable
                        else:
                            program['programexe'] = menuData['command']
                
                programs[program['name']] = program
                self.debug("Program added: %s" % program)
        
        if include_menu_only_entries:
            menuprograms = {}
            for menuKey,menuData in menuOnlyPrograms.iteritems():
                menuData = util.dictToCaseInsensitive(menuData)
                program = {}
                program['name'] = menuData['name']
                if 'command' in menuData and menuData['command']:
                    if menuData['command'].startswith('env WINEPREFIX=') or menuData['command'].startswith('wine '):
                        executable = self.__returnExecutableFromCommand(menuData['command'])
                        executable = util.wintounix(executable)
                        if self.__isValidProgramName(executable):
                            program['programcommand'] = menuData['command']
                            program['programexe'] = executable
                if 'icon' in menuData and menuData['icon']:
                    program['icon'] = menuData['icon']
                if 'comment' in menuData and menuData['comment']:
                    program['description'] = menuData['comment']
                menuprograms[menuData['name']] = program
            return (programs, menuprograms)
        else:
            return programs
    
    def getPrint(self):
        programs = self.list()
        if len(programs):
            current = 1
            for name in sorted(programs.keys(), key=str.lower):
                program = programs[name]
                print "Program:\t%s" % program['name']
                for key in ['Version:', 'Publisher:', 'Uninstall:', 'Description:', 'Icon:\t', 'ProgramIcon:', 'ProgramCommand:', 'ProgramExe:']:
                    if key.lower().split(':')[0] in program:
                        print "%s\t%s" % (key, program[key.lower().split(':')[0]])
                if current != len(programs.keys()):
                    print ""
                current += 1
        else:
            print "No installed programs in registry."
        
    def __returnValidProgramName(self, name, clean=True):
        if not clean:
            # Change Window's simple character replacements with real characters
            name = unicode(name)
            name = re.sub(r'(?i)\(tm\)', u'™', name)
            name = re.sub(r'(?i)\(c\)', u'©', name)
            name = re.sub(r'(?i)\(r\)', u'®', name)
        # Remove anything after and including first parenthesis if it includes the word "uninstall"
        # (some programs have stuff like "(uninstall only)" in their name, this strips it)
        name = re.sub(r'(?i) ?\([^\)]*?uninstall[^\(]*?\)','', name)
        # Remove any version number at the end of the name
        name = re.sub(' [vV]?\.?\d*?(\.)?(\d+\.?)*?$','', name)
        if clean:
            return name.lower()
        else:
            return name
    
    def __isValidProgramName(self, name, remove_help=True):
        badwords = self.IGNORE_PROGRAMS_CONTAINING
        if remove_help:
            badwords += self.DOCUMENT_MATCHERS
        name = name.lower()
        for badword in badwords:
            if badword in name:
                return False
        return True
    
    def __programNamesMatch(self, name1, name2):
        name1 = re.sub('(^play |\..{3}$)', '', self.__returnValidProgramName(name1))
        name2 = re.sub('(^play |\..{3}$)', '', self.__returnValidProgramName(name2))
        if name1 == name2 or \
           name1.startswith(name2) or name2.startswith(name1) or \
           name1.endswith(name2) or name2.endswith(name1):
            return True
        return False
    
    def __returnExecutableFromCommand(self, command):
        """ Remove C:\Windows\Command\Start.exe from command """
        command = re.sub(r'(?i)"?[a-z]\:\\+windows\\+command\\+start\.exe"? ?','', command)
        """ Isolate the first executable in the command """
        return re.sub(u'(^(env WINEPREFIX=".*?" )?wine "|(?<=[^^\n\r])".*$)', '', command)
    
    def setProgramOptions(self, key, name=None, icon=None, command=None, description=None, terminal=None, uninstall=None):
        regData = {}
        if name != None:
            regData['VineyardName'] = name
        if icon != None:
            regData['VineyardIcon'] = icon
        if command != None:
            regData['VineyardCommand'] = command
        if description != None:
            regData['VineyardDescription'] = description
        if terminal != None:
            regData['VineyardTerminal'] = terminal
        if uninstall != None:
            regData['VineyardUninstall'] = uninstall
        #print "Setting or adding programkey \"%s\": %s" % (key, regData)
        self.registry.set({'HKEY_LOCAL_MACHINE\\Software\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\%s' % key: regData})
            
    
    def createDesktopFiles(self):
        for programname, programdata in self.list().iteritems():
            if 'name' not in programdata:
                programdata['name'] = programname
            self.createDesktopFile(programdata, "/tmp")
    
    def createDesktopFile(self, program, filedir=None):
        if type(program) != type({}):
            program = self.list()[program]
        
        if filedir == None:
            filedir = os.path.realpath(os.path.expanduser("~/.local/share/applications/"))
        if not os.path.isdir(filedir):
            if subprocess.call(['mkdir', '--parent', filedir]) != 0:
                self.error("Error: Couldn't create \"%s\"" % filedir)
                return False
        
        programname = str(re.sub('[^\w,.\-+\(\)=\ ]', '?', program['name'].lower()))
        
        desktop = "[Desktop Entry]\nX-AppInstall-Package=%s\nX-AppInstall-Popcon=%s\nX-AppInstall-Section=main\n\nName=%s\n" % (programname, 0, program['name'])
        if 'programcommand' in program:
            desktop += "Exec=%s\n" % program['programcommand']
        elif 'programexe' in program:
            if os.path.realpath(os.environ['WINEPREFIX']) == os.path.realpath(util.getRealHome()+"/.wine"):
                desktop += "Exec=wine \"%s\"\n" % program['programexe']
            else:
                desktop += "Exec=env HOME=\"%s\" WINEPREFIX=\"%s\" wine \"%s\"\n" % (os.environ['HOME'], os.environ['WINEPREFIX'], program['programexe'])
        desktop += "Terminal=false\nType=Application\nStartupNotify=true\n"
        desktop += "Icon=%s\n" % self.getIcon(program)
        if 'description' in program:
            desktop += "Comment=%s\n" % program['description']
        if 'installlocation' in program:
            desktop += "Path=%s\n" % util.wintounix(program['installlocation'])
        desktop += "Categories=Wine;\nX-Ubuntu-Gettext-Domain=app-install-data\n\nX-Wine-Application=true"
        if 'uninstall' in program:
            desktop += "\nX-Wine-Uninstall=%s" % program['uninstall']
        desktopfilename = "%s/wine-application-%s.desktop" % (filedir, programname)
        desktopfile = open(desktopfilename, 'w')
        desktopfile.write(desktop)
        desktopfile.close()
        return desktopfilename
    
    def getIcon(self, program, force_update=False):
        if type(program) != type({}):
            program = self.list()[program]
        
        programname = self.__returnValidProgramName(program['name'])
        
        #self.debug("Looking for icon for \"%s\"" % program['name'])
        if 'icon' in program or 'programicon' in program or 'programexe' in program:
            iconpath = os.path.realpath("%s/../.icons/wine-application-icon-%s.png" % (os.environ['WINEPREFIX'], programname))
            if force_update or not os.path.exists(iconpath):
                if not os.path.isdir(os.path.dirname(iconpath)):
                    os.mkdir(os.path.dirname(iconpath))
                
                if 'icon' in program:
                    icon = program['icon']
                elif 'programicon' in program:
                    icon = program['programicon']
                    #self.debug("\tUsing programicon \"%s\" as icon." % program['programicon'])
                else:
                    icon = program['programexe']
                    #self.debug("\tUsing programexe \"%s\" as icon." % program['programexe'])
                
                if type(icon) != str and type(icon) != unicode:
                    self.critical("Icon has wrong value type %s for program \"%s\"." % (type(icon), program['name']))
                    return None
                
                if icon.lower().endswith('.exe') or icon.lower().endswith('.dll'):
                    # Help for specific programs
                    """for namematch in self.PROGRAM_OVERRIDES.keys():
                        if re.search(namematch, icon):
                            self.debug("\tName override for icon found, performing magic...")
                            icon = re.sub(namematch, self.PROGRAM_OVERRIDES[namematch], icon)"""
                    
                    #self.debug("\tCreating icon from exe \"%s\"..." % icon)
                    self.createIconFromExe(icon, iconpath)
                else:
                    iconpath = icon
                    #self.debug("\tCached icon found, using\"%s\"." % iconpath)
        else:
            #self.debug("\tNo info to find icon from, no icon found.")
            return None
        #self.debug("\tDone. Icon for \"%s\" is located at \"%s\"" % (program['name'], iconpath))
        return util.getIconPathFromName(iconpath)

    def createIconFromExe(self, exe, destination):
        if os.path.exists(exe):
            tmpfilename = util.tempname('vineyard-icon-extraction-', '.ico')
            tmpfile = open(tmpfilename, 'w')
            # Extract all icons from inputfile
            subprocess.Popen(["wrestool", exe, "--extract", "--type=group_icon"], stdout=tmpfile).wait()
            tmpfile.close()
            # If the extraction gave any fruit
            if os.stat(tmpfilename).st_size:
                icons = filter(len, subprocess.Popen(["icotool", "--list", tmpfilename], stdout=subprocess.PIPE).communicate()[0].split('\n'))
                # If the icon file has any icons
                if len(icons):
                    # Sort the icons by width and then by bit depth and pick the first (largest width and highest bit depth)
                    # More specific iconindex is width+bitdepth.index - f.x. 324.1
                    iconindex = sorted([ "%s%s.%s" % (i.split('=')[2].split()[0], i.split('=')[4].split()[0], i.split('=')[1].split()[0]) for i in icons ], reverse=True)[0]
                    # Extract and convert the best icon (found above)
                    return subprocess.call(["icotool", "-x", "--index", iconindex.split('.')[1], tmpfilename, "-o", destination])
            else:
                self.error("Error: Couldn't read icon(s) from file \"%s\"" % exe)
                return False
        else:
            return False

    def uninstall(self, program):
        programs = self.listFromReg()
        if program in programs:
            process = subprocess.Popen(["wine", programs[program]['UninstallString']], stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr, env=os.environ)
            return process.wait()
        else:
            return False
